"""Standard responses to client"""


def error_response_service(msg: str = ""):
    """Define a error json response to send a client.

    :param msg: A message indicating that the request has errors.
    """
    _data_response = {
        u'valid': False,
        u'msg': msg
    }
    return _data_response


def success_response_service(data: any = None, msg: str = ""):
    """Defines the structure of a response to a client request in JSON format.

    :param data: Information to send to client.
    :param msg: A message indicating that the request completed successfully.
    """
    _data_response = {
        u'valid': True,
        u'msg': msg,
        u'data': data
    }
    return _data_response



# Pytest
import pytest

# Werkzeug
from werkzeug.test import Client

# Retic
from retic import App as app, Router

PATHS = [
    ("/"),
]
PATHS_ROUTES = [
    ("/endpoint")
]
PATHS_SLASH = [
    ("/examples/")
]

CONTROLLERS = [
    # Check if the value is like as bytes, return string
    lambda req, res: res.ok({u'type': req.body.type, u'value': req.body.value if not isinstance(
        req.body.value, bytes) else req.body.value.decode("utf-8")})
]


@pytest.fixture
def app_without_client():
    """Clear the app"""
    app.clear()
    """Returns an app client with routes"""
    _router = Router()
    for _path in PATHS+PATHS_SLASH+PATHS_ROUTES:
        """define a new path using the response from a path definition"""
        _router \
            .get(_path, *CONTROLLERS) \
            .post(_path, *CONTROLLERS)
    app.use(_router)
    return app


@pytest.mark.lib
def test_config_attribute():
    """Clear the configuration"""
    app.config.clear()
    """Check that the variables aren't exists"""
    assert app.config.get("APP_LANG") == None
    assert app.config.get("APP_HOSTNAME") == None
    """we include a basic setting to app"""
    app.config.set("APP_LANG", "en_US")
    app.config.set("APP_HOSTNAME", "localhost")
    """Check if this value was saved in the object"""
    assert app.config.get("APP_LANG") == "en_US", \
        "The value from the configuration item is different to value saved"
    assert app.config.get("APP_HOSTNAME") == "localhost", \
        "The value from the configuration item is different to value saved"


@pytest.mark.lib
def test_config_attribute():
    """Clear the configuration"""
    app.config.clear()
    """We don't include a basic setting to app and check if this value wasn't saved in the object"""
    assert app.config.get("APP_LANG") == None


@pytest.mark.lib
def test_config_from_object():
    """Clear the configuration"""
    app.config.clear()
    """Check that the variables aren't exists"""
    assert app.config.get("APP_LANG") == None
    """Set the settings from an object"""
    app.config.from_object({u'APP_LANG': "en_US"})
    """check if this value exists"""
    assert app.config.get("APP_LANG") == "en_US", \
        "The value from the configuration item is different to value saved"

@pytest.mark.lib
def test_config_clear():
    """Set the settings from an object"""
    app.config.from_object({u'APP_LANG': "en_US"})
    """check if this value exists"""
    assert app.config.get("APP_LANG") == "en_US", \
        "The value from the configuration item is different to value saved"
    """Clear the configuration"""
    app.config.clear()
    """Check that the variables aren't exists"""
    assert app.config.get("APP_LANG") == None


"""Test about main App"""


@pytest.mark.lib_hooks
@pytest.mark.parametrize("path", PATHS_ROUTES)
def test_request_clear_app(app_without_client, path):
    _app = Client(app_without_client.application)
    """get a request when the app has routes"""
    app_iter, status, headers = _app.get(path)
    assert status.upper() == "200 OK"
    """we clear the information of App"""
    app_without_client.clear()
    """get a request when the app hasn't routes"""
    app_iter, status, headers = _app.get(path)
    assert status.upper() == "404 NOT FOUND"



# Utils
from retic.services.exceptions import get_file_error_exception
from retic.services.core.general import path_regexp


class Layer(object):
    def __init__(self, path: str, options: dict, fn: any):
        """Initial instance of the Layer Class

        :param path: Route to searching in the layer
        :param options: Options for the regexp expression
        :param fn: Route dispatch function for the request"""
        self.handle = fn
        self.keys = []
        self.name = fn.__name__ if fn.__name__ else '<anonymous>'
        self.params = self.path = self.route = None
        self.regexp = path_regexp(path, self.keys, options)

    def handle_request(self, req, res, next):  # dispatch
        """Return a handle request for specific route"""
        try:
            return self.handle(req, res, next)
        except Exception as e:
            get_file_error_exception(3)
            raise RuntimeError(
                "error: Some controller has a problem, check it please. {0}".format(
                    str(e)
                )
            )

    def match(self, path: str):  # self
        """Search to specific layer for a path

        :param path: Route to searching in the layer
        :return: Specific layer that was matched with the path"""
        _match = self.regexp.match(path)
        if not _match:
            self.path = self.params = None
            return None
        self.path = _match.string
        self.params = _match.groupdict()
        return self



# Gettext
import gettext

# Retic
from retic import cfg

DOMAIN = cfg.APP_LANG_DOMAIN
LOCALEDIR = cfg.APP_LANG_LOCALEDIR
LANGUAGES = cfg.APP_LANG_LANGUAGES
FALLBACK = cfg.APP_LANG_FALLBACK
_t = None

if LOCALEDIR:
    _exists = gettext.find(DOMAIN, LOCALEDIR, LANGUAGES, all=False)
    if not _exists:
        raise FileNotFoundError(
            "{0} doesn't have a valid .mo translate file, check if you have a .po files"
        ).format(LOCALEDIR)

    t = gettext.translation(
        domain=DOMAIN,
        localedir=LOCALEDIR,
        languages=LANGUAGES,
        fallback=FALLBACK
    )
    _t = t.gettext



# Time
import time


def sleep(seconds: int = 0):
    """The Python sleep () time method suspends execution for the 
    given number of seconds. The argument can be a floating point 
    number to indicate a more accurate sleep time.

    :param seconds: Total of seconds that the program will be sleeping
    """
    time.sleep(seconds)



# Utils
from retic.services.core.general import get_number_parameters


class Route(object):
    def __init__(self, path: str):
        """Initial instance of the Route Class

        :param path: Route to searching in the layer"""
        self.methods = None
        self.stack = []
        self.path = path

    def dispatch(self, req, res, next):  # function
        """Interface beetwen the *controller function** and **route**

        :param req: Request is used to describe an request to a server.
        :param res: Represents a response from a web request.
        :param next: It must call next() to pass control to the next middleware function
        """
        _handles = next.stack
        """Check if the next function exists"""
        if not _handles:
            return
        # assert _handles, "error: Next function is invalid, the function has not more controllers"

        _fn = _handles.pop(0).handle
        _num_param = get_number_parameters(_fn)

        if _num_param == 3:
            return _fn(req, res, next.next)
        elif _num_param == 2:
            return _fn(req, res)
        raise Exception("error: The controllers parameters are invalid")



# Repath
from repath import pattern

# Inspect
import inspect

# Re
import re


def path_regexp(path, keys, options):
    """ Return a regular expression function for a specific route

    :param path: Route to searching in the layer    
    :param keys: List of keys from the route
    :param options: Options for the regexp expression    
    :return: A regular expression pattern string
    """
    _reg_exp = re.compile(pattern(path, **options))
    for _key in _reg_exp.groupindex:
        keys.append({
            u"name": _key,
            u"optional": False
        })
    return _reg_exp


def get_number_parameters(func):
    """Returns the number of parameters of a specific function."""
    return len(inspect.signature(func).parameters)


def get_body_request(app_iter):
    """Generate a string from response of the app requests

    :param app_iter: Response from the request to api
    """
    return (b''.join(app_iter).splitlines()[0]).decode("utf-8")


# Traceback
import traceback

# Sys
import sys

# Utils
from retic.services.exceptions import get_file_error_exception


class HttpMethod(object):
    def __init__(self, router, method: str, route: any):
        """Initial instance of the HttpMethod Class.

        :param router: Instance of the Router Class, this is use when I
            need to use router.get(...).get(...).post(...)
        :param method: HTTP Request Method from the client request
        :param route: Route function from Router class
        """
        self.router = router
        self.method = method
        self.route = route

    def default(self, *args):  # self
        """Create a new route in base a path and controllers from the routes config

        :param method: HTTP Request Method from the client request
        :param route: Route function from Router class
        """
        if len(args) < 2:
            raise ValueError(
                "error: The route has the next format: METHOD(path, [...handlers functions])"
            )
        path = args[0]
        if type(path) is not str:
            raise TypeError("error: The path type must be a string format")
        self.route(path, self.method.upper(), args[1:])
        return self.router



# Werkzeug
from werkzeug.wrappers import Response
from werkzeug.routing import RequestRedirect

# Httpmethods
from httpmethods import get_status_by_code

# Utils
from retic.services.general.json import jsonify


class Response(Response):

    def _config(
        self,
        environ,
        start_response,
        set_response
    ):
        """Config instance of the Responses Class

        :param environ: Represents a request from a web request.
        :param start_response: Represents a response from a web request.
        :param set_response: Response function for the request.
        """
        self._environ = environ
        self._start_response = start_response
        self._set_response = set_response
        # set default headers
        self.set_headers('content-type', "application/json")
        return self

    def bad_request(self, content: any = ""):  # Response
        """This method response a *400 Bad Request* response to the client,
        this indicate that the request is invalid.

        This generally means that the request contained invalid parameters
        or headers, or that you tried to do something that your application
        logic does not support.

        :param content: Information to send the client, a message, a dict, etc.
        If it doesn't exist, it sends a status message from the status code.
        """
        return self._send_by_status(400, content)

    def forbidden(self, content: any = ""):  # Response
        """This method is used to send a *403 Forbidden* response to the client,
        indicating that a request is not authorized.

        This generally means that the user agent tried to do something that they
        were not authorized to do, such as changing another user's password.

        :param content: Information to send the client, a message, a dict, etc.
        If it doesn't exist, it sends a status message from the status code.
        """
        return self._send_by_status(403, content)

    def not_found(self, content: any = ""):  # Response
        """This method is used to send a *404 Not found* response.

        When called manually from your application code, this method is normally used to
        indicate that the user agent tried to find, update, or delete something that doesn't exist.

        :param content: Information to send the client, a message, a dict, etc.
        If it doesn't exist, it sends a status message from the status code.
        """
        return self._send_by_status(404, content)

    def ok(self, content: dict = None):  # Response
        """This method is used to send a *200 OK* response to the client.

        :param content: Information to send the client, a message, a dict, etc.
        If it doesn't exist, it sends a status message from the status code.
        """
        return self._send_by_status(200, content)

    def server_error(self, content: any = ""):
        """This method is used to send a *500 Server Error* response to the client,
        indicating that some kind of server error occurred.

        :param content: Information to send the client, a message, a dict, etc.
        If it doesn't exist, it sends a status message from the status code.
        """
        return self._send_by_status(500, content)

    def send(self, content: any = None):  # Response
        """Send a string response in a non-JSON format (XML, CSV, plain text, etc.).

        This method is used in the underlying implementation of most other terminal response methods.

        :param content: Information to send the client, a message, a dict, etc.
        If it doesn't exist, it sends a status message from the status code.
        """
        return self._send_string(content if content else {})

    def set_headers(self, headers: dict, value: str = None):  # self
        """Sets the specified response header to the specified value.

        Alternatively, you can pass a single object argument (headers)
        to set multiple header fields at once, where the keys are the names
        of the header fields and the corresponding values ​​are the desired values.

        :param headers: It can be of type ``dict``, to represent a headers object 
        that are added to the current headers. If it is of type ``str`` it will 
        be used to access a specific header. Any other format may cause an error exception.

        :param value: Value to assign to the specified header. By default it has a 
        value of ``None``.
        """
        if type(headers) is dict and value is None:
            self.headers = {**self.headers, **headers}
        elif type(headers) is str:
            self.headers[headers] = value
        else:
            raise TypeError(
                "error: Headers have an invalid type. You can use dict or str."
            )
        return self

    def set_status(self, code: int):  # self
        """Set the status code for this response.

        :param code: Number representing the status code of the HTTP response
        """
        self.status_code = code
        return self

    def redirect(self, new_url: str):  # Response
        """Redirect to another url with the actual request
        
        :param new_url: URL to redirect.
        """
        _result = RequestRedirect(new_url=new_url).get_response(self._environ)
        return self._set_response(_result(self._environ, self._start_response))

    def _send_string(self, content: str = ""):  # Response
        """Send a response to http requests.

        :param content: Information to send the client, a message, a dict, etc.
        If it doesn't exist, it sends a status message from the status code.
        """
        self.set_data(jsonify(content))
        return self._set_response(self(self._environ, self._start_response))

    def _send_by_status(self, status, content):
        return self.set_status(status)._send_string(content or get_status_by_code(status))

    def favicon(self, req, res):
        return self.ok()



# Environs
from environs import Env

# Utils
import retic.services.core.msgfmt as msgfmt

env = Env()
env.read_env()

# TODO: Implement gettext libs
def build_api(env):
    # load files
    msgfmt.make("retic\locale", "base")
    _locale = env("APP_LANG_LOCALEDIR", None)
    if not _locale:
        return
    msgfmt.make(_locale, env.str("APP_LANG_DOMAIN", "base"))


# build_api(env)



# Werkzeug
from werkzeug.test import Client

# Pytest
import pytest

# Retic
from retic import App as app, Router

# Utils
from retic.services.core.general import get_body_request

PATHS = [
    ("/withoutress")
]

CONTROLLERS = [
    lambda req, res: print("REST api Python example 🐍")
]


@pytest.fixture
def app_client():
    """Clear the app"""
    app.clear()
    """Returns an app client without routes"""
    return Client(app.application)


@pytest.fixture
def app_routes():
    """Clear the app"""
    app.clear()
    """Returns an app client with routes"""
    _router = Router()
    for _path in PATHS:
        """define a new path using the response from a path definition"""
        _router \
            .get(_path, *CONTROLLERS) \
            .get("/", *CONTROLLERS)
    app.use(_router)
    return Client(app.application)


@pytest.mark.lib_hooks
@pytest.mark.parametrize("path", PATHS)
def test_response_without_method(app_client, path):
    """we include a valid route and controllers"""
    app_iter, status, headers = app_client.get(path)
    assert status.upper() == '404 NOT FOUND', "A status 404 is necesary, but a status {} was got from the request".format(
        status)


@pytest.mark.lib_hooks
@pytest.mark.parametrize("path", PATHS)
def test_response_without_method_routes(app_routes, path):
    """we include a valid route and controllers"""
    app_iter, status, headers = app_routes.get(path)
    assert status.upper() == '200 OK', "A status 200 is necesary, but a status {} was got from the request".format(
        status)
    assert get_body_request(
        app_iter) == '200 OK', "The default from the api when this one doesn't have routes is different to documentation"



# Werkzeug
from werkzeug.wrappers import Request

# Utils
from retic.services.general.json import parse


class Body(object):
    """Class for the body from a request"""

    def __init__(self, type, value):
        """Define the type of the object"""
        self.type: dict = type
        """Value of the object"""
        self.value: dict = value

    @property
    def type(self):
        return self.__type

    @type.setter
    def type(self, value):
        self.__type = value

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, value):
        self.__value = value


class Request(Request):
    """Initial instance of the Request Class"""

    @property
    def retic(self):
        return self.__retic

    @retic.setter
    def retic(self, value):
        if hasattr(self, "retic"):
            raise TypeError(
                "error: You can't assign the settings of this ways. if you want to assign from an object please use to *req.retic.clear()* function"
            )
        self.__retic = value

    def _config(self):
        """Set another attributes to Instance"""
        self.body: Body = self._get_body()
        self.retic = self.params = {}
        return self

    def param(self, key: str, default_value: any = None, callback=None):
        """Returns the value of the parameter with the specified name.

        req.param(...) finds in the URL path, body, and query string (in that order) 
        for the specified parameter. If no parameter value exists anywhere in the 
        request with the given name, it returns None or the optional default value if specified.

        :param key: Name of the parameter to find
        :param default_value: Value of the parameter if this one doesn't exist
        :param callback: Function that is executed after getting the value        
        """
        if key in self.params:
            _value = self.params.get(key)
        elif key in self.body.value:
            _value = self.body.value.get(key)
        else:
            _value = self.args.get(key, default_value)
        if not callback:
            return _value
        return callback(_value)

    def set(self, key: str, value: any = None):
        """Set a value in the requests (req).

        Please note that names are not case sensitive.

        :param key: Name of the variable to set
        :param value: Value of the variable
        """
        return self.retic.setdefault(key.lower(), value)

    def get(self, key: str, default_value: any = None):
        """Returns the value of an object in retic with a specific name. 
        Please note that names are not case sensitive. 

        If the value does not exist in the request, it returns None or 
        the default value specified by default.

        :param key: Name of the variable to find
        :param default_value: Value of the parameter if this one doesn't exist
        """
        return self.retic.get(key.lower(), default_value)

    def all_params(self):
        """Returns the value of all the parameters sent in the request,
        combined into a single dictionary.

        It includes parameters parsed from the URL path, the request body,
        and the query string, retic dict in that order."""
        return {**self.params, **self.body.value, **self.args, **self.retic}

    def _get_body(self):
        """Get the body for a request from to client. If this one is not exists, return

        ``{ type: "undefiend", value: "undefiend" }``

        The response takes possibly four keys: json, form, text, and raw. 

        For example if the request body is a json object, you will get:

        ``{ type: "json", value: json_object }``

        Also, if a exception is detected, you will get

        ``{ type: "error", value: error_object }``
        """
        _type = ''
        _value = ''
        _content_type = self.headers.get('content-type', type=str)
        try:
            if not _content_type:
                _type = 'undefiend'
                _value = 'undefiend'
            elif _content_type.startswith('application/json'):
                _type = "json"
                _value = parse(self.get_data())
            elif _content_type.startswith('multipart/form-data') \
                    or _content_type.startswith('application/x-www-form-urlencoded'):
                _type = 'form'
                _value = self.form
            elif _content_type.startswith('text/'):
                _type = "text"
                _value = self.get_data().decode("utf-8")
            else:
                _type = 'raw'
                _value = self.get_data()
            return Body(
                type=_type,
                value=_value,
            )
        except Exception as e:
            return Body(
                type='error',
                value=e,
            )



#! /usr/bin/env python3
# Written by Martin v. Löwis <loewis@informatik.hu-berlin.de>

"""Generate binary message catalog from textual translation description.

This program converts a textual Uniforum-style message catalog (.po file) into
a binary GNU catalog (.mo file).  This is essentially the same function as the
GNU msgfmt program, however, it is a simpler implementation.  Currently it
does not handle plural forms but it does handle message contexts.

Usage: msgfmt.py [OPTIONS] filename.po

Options:
    -o file
    --output-file=file
        Specify the output file to write to.  If omitted, output will go to a
        file named filename.mo (based off the input file name).

    -h
    --help
        Print this message and exit.

    -V
    --version
        Display version information and exit.
"""

import os
import sys
import ast
import getopt
import struct
import array
from email.parser import HeaderParser

__version__ = "1.2"

MESSAGES = {}


def usage(code, msg=''):
    print(__doc__, file=sys.stderr)
    if msg:
        print(msg, file=sys.stderr)
    sys.exit(code)


def add(ctxt, id, str, fuzzy):
    "Add a non-fuzzy translation to the dictionary."
    global MESSAGES
    if not fuzzy and str:
        if ctxt is None:
            MESSAGES[id] = str
        else:
            MESSAGES[b"%b\x04%b" % (ctxt, id)] = str


def generate():
    "Return the generated output."
    global MESSAGES
    # the keys are sorted in the .mo file
    keys = sorted(MESSAGES.keys())
    offsets = []
    ids = strs = b''
    for id in keys:
        # For each string, we need size and file offset.  Each string is NUL
        # terminated; the NUL does not count into the size.
        offsets.append((len(ids), len(id), len(strs), len(MESSAGES[id])))
        ids += id + b'\0'
        strs += MESSAGES[id] + b'\0'
    output = ''
    # The header is 7 32-bit unsigned integers.  We don't use hash tables, so
    # the keys start right after the index tables.
    # translated string.
    keystart = 7*4+16*len(keys)
    # and the values start after the keys
    valuestart = keystart + len(ids)
    koffsets = []
    voffsets = []
    # The string table first has the list of keys, then the list of values.
    # Each entry has first the size of the string, then the file offset.
    for o1, l1, o2, l2 in offsets:
        koffsets += [l1, o1+keystart]
        voffsets += [l2, o2+valuestart]
    offsets = koffsets + voffsets
    output = struct.pack("Iiiiiii",
                         0x950412de,       # Magic
                         0,                 # Version
                         len(keys),         # # of entries
                         7*4,               # start of key index
                         7*4+len(keys)*8,   # start of value index
                         0, 0)              # size and offset of hash table
    output += array.array("i", offsets).tobytes()
    output += ids
    output += strs
    return output


def make(folder, domain):
    global MESSAGES
    msgid = msgstr= b''
    ID = 1
    STR = 2
    CTXT = 3

    _files = []
    for root, dirs, files in os.walk(folder):
        outfile = "{0}\{1}.mo".format(
            root, domain)
        _items = []
        for file in files:
            if file.endswith(".po"):
                _items.append(os.path.join(root, file))
        if _items:
            _files.append({
                'outfile': outfile,
                'items': _items
            })
    for file in _files:
        MESSAGES.clear()
        lines = []
        outfile = file.get("outfile")
        items = file.get('items')
        for filename in items:
            # Compute .mo name from .po name and arguments
            if filename.endswith('.po'):
                infile = filename
            else:
                infile = filename + '.po'
            if outfile is None:
                outfile = os.path.splitext(infile)[0] + '.mo'

            try:
                with open(infile, 'rb') as f:
                    lines = lines+f.readlines()
            except IOError as msg:
                print(msg, file=sys.stderr)
                sys.exit(1)

        section = msgctxt = None
        fuzzy = 0

        # Start off assuming Latin-1, so everything decodes without failure,
        # until we know the exact encoding
        encoding = 'latin-1'

        # Parse the catalog
        lno = 0
        for l in lines:
            l = l.decode(encoding)
            lno += 1
            # If we get a comment line after a msgstr, this is a new entry
            if l[0] == '#' and section == STR:
                add(msgctxt, msgid, msgstr, fuzzy)
                section = msgctxt = None
                fuzzy = 0
            # Record a fuzzy mark
            if l[:2] == '#,' and 'fuzzy' in l:
                fuzzy = 1
            # Skip comments
            if l[0] == '#':
                continue
            # Now we are in a msgid or msgctxt section, output previous section
            if l.startswith('msgctxt'):
                if section == STR:
                    add(msgctxt, msgid, msgstr, fuzzy)
                section = CTXT
                l = l[7:]
                msgctxt = b''
            elif l.startswith('msgid') and not l.startswith('msgid_plural'):
                if section == STR:
                    add(msgctxt, msgid, msgstr, fuzzy)
                    if not msgid:
                        # See whether there is an encoding declaration
                        p = HeaderParser()
                        charset = p.parsestr(msgstr.decode(
                            encoding)).get_content_charset()
                        if charset:
                            encoding = charset
                section = ID
                l = l[5:]
                msgid = msgstr = b''
                is_plural = False
            # This is a message with plural forms
            elif l.startswith('msgid_plural'):
                if section != ID:
                    print('msgid_plural not preceded by msgid on %s:%d' % (infile, lno),
                          file=sys.stderr)
                    sys.exit(1)
                l = l[12:]
                msgid += b'\0'  # separator of singular and plural
                is_plural = True
            # Now we are in a msgstr section
            elif l.startswith('msgstr'):
                section = STR
                if l.startswith('msgstr['):
                    if not is_plural:
                        print('plural without msgid_plural on %s:%d' % (infile, lno),
                              file=sys.stderr)
                        sys.exit(1)
                    l = l.split(']', 1)[1]
                    if msgstr:
                        msgstr += b'\0'  # Separator of the various plural forms
                else:
                    if is_plural:
                        print('indexed msgstr required for plural on  %s:%d' % (infile, lno),
                              file=sys.stderr)
                        sys.exit(1)
                    l = l[6:]
            # Skip empty lines
            l = l.strip()
            if not l:
                continue
            l = ast.literal_eval(l)
            if section == CTXT:
                msgctxt += l.encode(encoding)
            elif section == ID:
                msgid += l.encode(encoding)
            elif section == STR:
                msgstr += l.encode(encoding)
            else:
                print('Syntax error on %s:%d' % (infile, lno),
                      'before:', file=sys.stderr)
                print(l, file=sys.stderr)
                sys.exit(1)
        # Add last entry
        if section == STR:
            add(msgctxt, msgid, msgstr, fuzzy)

        # Compute output
        output = generate()

        try:
            with open(outfile, "wb") as f:
                f.write(output)
        except IOError as msg:
            print(msg, file=sys.stderr)



# Inspect
import inspect

# Os
import os


def get_file_error_exception(position=2):
    """Returns a file path of a file that created an invocation function"""
    _calframe = inspect.getouterframes(inspect.currentframe(), 2)
    return os.path.abspath(inspect.getfile(_calframe[position][0]))



class Next(object):
    def __init__(self, req: any, res: any, layer: any, middleware: list):
        """Initial instance of the Next Class

        :param req: Request is used to describe an request to a server.
        :param res: Represents a response from a web request.
        :param layer: Layer that contains a route instance for a specific path
        """
        self.layer = layer
        """Add all middlewares available"""
        self.stack = middleware.copy()+layer.route.stack.copy()
        self.req = req
        self.res = res

    def next(self):
        """It must call next() to pass control to the next middleware function"""
        self.layer.handle_request(self.req, self.res, self)



# Re
import re

# Urllib
import urllib.parse

# Unicodedata
import unicodedata


def urlencode(url: str):
    """This function is convenient when encoding a string to 
    use in a query part of a URL, as a convenient way to pass 
    variables to the next page.

    :param url: URL to encode.
    """
    key = urllib.parse.quote_plus(url)
    return key


def slugify(text: str = ""):
    """Simplifies ugly strings into something URL-friendly.

    :param text: Text to simplifies in ``str`` type    
    """
    # Reference: https://github.com/mikaelho/docgen/blob/master/docgen.py
    # "[Some] _ Article's Title--"
    # "[some] _ article's title--"
    text = str(text).lower()
    # "[some] _ article's_title--"
    # "[some]___article's_title__"
    for c in [' ', '-', '.', '/']:
        text = text.replace(c, '_')
    # "[some]___article's_title__"
    # "some___articles_title__"
    text = re.sub('\W', '', text)
    # "some___articles_title__"
    # "some   articles title  "
    text = text.replace('_', ' ')

    # "some   articles title  "
    # "some articles title "
    text = re.sub('\s+', ' ', text)

    # "some articles title "
    # "some articles title"
    text = text.strip()

    # "some articles title"
    # "some-articles-title"
    text = text.replace(' ', '-')
    # delete acents
    text = unicodedata.normalize("NFD", text)
    text = text.encode("utf8").decode("ascii", "ignore")
    text = urlencode(text)
    return text



# Werkzeug
from werkzeug.routing import RequestRedirect

# Httpmethods
from httpmethods import get_http_methods

# Inspect
from inspect import getfile

# Traceback
import traceback

# Sys
import sys

# Os
import os

# Retic
from retic.lib.hooks.responses import Request, Response, Next
from retic.lib.router.httpmethod import HttpMethod
from retic.lib.router.route import Route
from retic.lib.router.layer import Layer


class Router(object):
    def __init__(self, strict_slashes: bool = True):
        """Initial instance of the Router Class

        :param strict_slashe: If a rule ends with a forward slash but the matching URL does not, redirect to the ending no forward slash URL.
        """
        self.__dict__ = {
            key: HttpMethod(self, key, self.route).default for key in get_http_methods()
        }
        self.name = "router"
        self.methods = {key: [] for key in get_http_methods(True)}
        self.args = self.params = []
        self.result = None
        self.rules = {
            "strict_slashes": strict_slashes
        }
        self.middlewares = []

    @property
    def middlewares(self):
        return self.__middlewares

    @middlewares.setter
    def middlewares(self, value):
        self.__middlewares = value

    @property
    def result(self):
        return self.__result

    @result.setter
    def result(self, value):
        self.__result = value

    def main(self, environ: dict, start_response: dict):  # []
        """Router main for handle all routes

        :param environ: Request is used to describe an request to a server.
        :param start_response: Represents a response from a web request."""
        try:
            _request = Request(environ)._config()
            _response = Response()._config(environ, start_response, self._set_response)
            self._endpoint(_request, _response)
            _result = self._response_request(_response, self.result)
            self.result = None
            return _result
        except Exception as e:
            return Response(str(e))(environ, start_response)

    def use(self, fn):
        """Use to add a middleware that is execute for all routes

        :param fn: Route dispatch function for the request
        """
        """Set middle to routes"""
        self.middlewares.append(
            self._set_handler_to_layer(fn)
        )

    def route(self, path: str, method: str, handlers: list):  # Route
        """Create a Route instance and adds this route to specific controller
        of a HTTP method in a Router.

        :param path: Route to searching in the layer
        :param method: HTTP Request Method from the client request
        :param handlers: List of controllers functions from a routes files
        """
        _route = Route(path)
        for _handler in handlers:
            """Set handler to route"""
            _route.stack.append(
                self._set_handler_to_layer(_handler)
            )
        _layer = Layer(
            path,
            {
                u"strict": self.rules.get('strict_slashes', False),
                u"end": True
            },
            _route.dispatch
        )
        _layer.route = _route
        self.methods[method].append(_layer)
        return _route

    def _set_handler_to_layer(self, handler):
        assert handler, "error: The route has the next format: METHOD(path, [...handlers functions])"
        assert callable(
            handler), "error: The handler or middleware is not a valid function"
        _layer = Layer(
            "/",
            {
                u"strict": False,
                u"end": True
            },
            handler
        )
        return _layer

    def _set_response(self, item: dict):
        """Response function for the request."""
        self.result = item
        return self.result

    def _response_request(self, res, result):
        """Response to a client request. If response was not specific, return
        status 200 and message 200 for default

        :param res: Represents a response from a web request.
        :param result: Instance of the object with the werkzeug response
        """
        if result:
            return result
        return res.ok()

    def _endpoint(self, req: Request, res: Response):  # []
        """This function handle any request from a client, search in the Route List

        :param req: Request is used to describe an request to a server.
        :param res: Represents a response from a web request."""
        try:
            # validate if path contains slash and it isn't a "/" path
            if self.rules.get('strict_slashes', False) \
                    and len(req.path) > 1 \
                    and "/" in req.path[-1]:
                raise ValueError(req.path[:-1])

            # search the specific method
            _method = self.methods[req.method]
            if not _method:
                raise KeyError(
                    "error: The HTTP method {0} doesn't exist".format(
                        req.method)
                )

            # search the layer for this method
            _layer: Layer = self._search_layer(req.path, _method)

            if not _layer:
                raise KeyError(
                    "error: The HTTP method {0} doesn't exist".format(
                        req.method)
                )

            # search the first handle for this one
            _has_method = self._handles_method(_layer)

            assert _has_method, "error: The route has the next format: METHOD(path, [...handlers functions])"

            # set the params to request
            req.params = _layer.params
            # return the handle logic
            return _layer.handle_request(req, res, Next(req, res, _layer, self.middlewares))
        except ValueError as e:
            return res.redirect(str(e))
        except KeyError as e:
            return res.not_found(str(e))
        except Exception as e:
            traceback.print_exc(file=sys.stdout)
            return res.bad_request(str(e))

    def _search_layer(self, path: str, method: str):  # Layer
        """Search a specific Layer using a path and method.

        :param path: Route to searching in the layer
        :param method: HTTP Request Method from the client request
        :return: None if the layer doesn't exists, and a Layer Object
        if the layer is found."""
        _match = None
        for _layer in method:
            _match = _layer.match(path)
            if _match:
                break
        return _match

    def _handles_method(self, layer: Layer):
        """Check if the handle from the Layer is a controller function."""
        return True if layer.route.stack[0].handle else False



# Pytest
import pytest

# Retic
from retic.lib.router import httpmethod, router

PATHS = [
    ("/withoutres"),
    ("/")
]

CONTROLLERS = [
    lambda req, res: print("REST api Python example 🐍")
]


@pytest.fixture
def http_method():
    """Returns a HttpMethod"""
    _router = router.Router()
    _method = 'GET'
    return httpmethod.HttpMethod(_router, _method, _router.route)


@pytest.mark.lib_router
@pytest.mark.parametrize("path", PATHS)
def test_default(http_method: httpmethod.HttpMethod, path):
    """we include a valid route and controllers"""
    _default = http_method.default(
        path,
        *CONTROLLERS
    )
    assert _default == http_method.router, "test failed"


@pytest.mark.lib_router
@pytest.mark.parametrize("path", PATHS)
def test_default_exception(http_method: httpmethod.HttpMethod, path):
    """we don't include a valid route, because the handles aren't passed"""
    with pytest.raises(ValueError) as excinfo:
        _default = http_method.default(
            path
        )
    assert "error: The route has the next format: METHOD(path, [...handlers functions])" == str(
        excinfo.value)


@pytest.mark.lib_router
@pytest.mark.parametrize("path", PATHS)
def test_default_type_error(http_method: httpmethod.HttpMethod, path):
    """we don't include a valid path, because is necesary a string for the path"""
    with pytest.raises(TypeError) as excinfo:
        _default = http_method.default(
            123,
            *CONTROLLERS
        )
    assert "error: The path type must be a string format" == str(
        excinfo.value
    )



# Werkzeug
from werkzeug.test import Client

# Pytest
import pytest

# Retic
from retic import App as app, Router
from retic.lib.hooks.system.env import env
from retic.services.core.general import get_body_request
from retic.services.general.json import parse


PATHS = [
    ("/"),
]
PATHS_ROUTES = [
    ("/endpoint")
]
PATHS_SLASH = [
    ("/examples/")
]

CONTROLLERS = [
    # Check if the value is like as bytes, return string
    lambda req, res: res.ok({u'type': req.body.type, u'value': req.body.value if not isinstance(
        req.body.value, bytes) else req.body.value.decode("utf-8")})
]


@pytest.fixture
def app_routes():
    """Clear the app"""
    app.clear()
    """Returns an app client with routes"""
    _router = Router()
    for _path in PATHS+PATHS_SLASH+PATHS_ROUTES:
        """define a new path using the response from a path definition"""
        _router \
            .get(_path, *CONTROLLERS) \
            .post(_path, *CONTROLLERS)
    app.use(_router)
    return Client(app.application)


@pytest.fixture
def app_without_client():
    """Clear the app"""
    app.clear()
    """Returns an app client with routes"""
    _router = Router()
    for _path in PATHS+PATHS_SLASH+PATHS_ROUTES:
        """define a new path using the response from a path definition"""
        _router \
            .get(_path, *CONTROLLERS) \
            .post(_path, *CONTROLLERS)
    app.use(_router)
    return app


"""Test about Body"""


@pytest.mark.lib_hooks
@pytest.mark.parametrize("path", PATHS)
def test_request_without_body(app_routes, path):
    """we include a valid route and controllers"""
    app_iter, status, headers = app_routes.get(path)
    _body = parse(get_body_request(app_iter))
    assert status.upper() == "200 OK"
    assert _body.get("type") == "undefiend"
    assert _body.get("value") == "undefiend"


@pytest.mark.lib_hooks
@pytest.mark.parametrize("path", PATHS)
def test_request_with_body_json(app_routes, path):
    """we include a valid route and controllers"""
    app_iter, status, headers = app_routes.post(path, json={'text': 'example'})
    _body = parse(get_body_request(app_iter))
    assert status.upper() == "200 OK"
    assert _body.get("type") == "json"
    assert _body.get("value") == {'text': 'example'}


@pytest.mark.lib_hooks
@pytest.mark.parametrize("path", PATHS)
def test_request_with_body_form(app_routes, path):
    """we include a valid route and controllers"""
    app_iter, status, headers = app_routes.post(path, data={'text': 'example'})
    _body = parse(get_body_request(app_iter))
    assert status.upper() == "200 OK"
    assert _body.get("type") == "form"
    assert _body.get("value") == {'text': 'example'}


@pytest.mark.lib_hooks
@pytest.mark.parametrize("path", PATHS)
def test_request_with_body_text(app_routes, path):
    """we include a valid route and controllers"""
    app_iter, status, headers = app_routes.post(
        path,
        data="{'text': 'example'}",
        content_type='text/plain'
    )
    _body = parse(get_body_request(app_iter))
    assert status.upper() == "200 OK"
    assert _body.get("type") == "text"
    assert _body.get("value") == "{'text': 'example'}"


@pytest.mark.lib_hooks
@pytest.mark.parametrize("path", PATHS)
def test_request_with_body_raw(app_routes, path):
    """we include a valid route and controllers"""
    app_iter, status, headers = app_routes.post(
        path,
        data="{'text': 'example'}",
        content_type='application/pdf'
    )
    _body = parse(get_body_request(app_iter))
    assert status.upper() == "200 OK"
    assert _body.get("type") == "raw"
    assert _body.get("value") == "{'text': 'example'}"


"""Test about Redirects"""


@pytest.mark.lib_hooks
@pytest.mark.parametrize("path", PATHS_SLASH)
def test_request_with_slash(app_routes, path):
    """we include a valid route and controllers"""
    app_iter, status, headers = app_routes.get(path)
    """Redirect to path without slash in the final"""
    _body = get_body_request(app_iter)
    assert status.upper() == "308 PERMANENT REDIRECT"



# Werkzeug
from werkzeug.serving import run_simple

# Environs
from environs import Env

# Retic
from retic.lib.router import Router

APP_HOSTNAME = "127.0.0.1"
APP_PORT = 1801


class App(object):
    def __init__(self, env):
        self.router: Router = None
        self.apps = {}
        self.env: Env = env
        self.config: Config = Config(env)

    @property
    def config(self):
        return self.__config

    @config.setter
    def config(self, value):
        """If the object exists you can't replace his value.

        Visit to documentation if you want clear the settings with the function
        ``app.config.clear()``

        Visit to documentation if you want set the settings from a onject with the funciont
        ``app.config.from_object()``
        """
        if hasattr(self, "config"):
            raise TypeError(
                "error: You can't assign the settings of this ways. if you want to assign from an object please use to *app.config.from_object()* function"
            )
        self.__config = value

    @property
    def router(self):
        return self.__router

    @router.setter
    def router(self, value):
        self.__router = value

    def application(self, environ, start_response):
        """Application for send the response returned by the application to the client

        :param environ: Request is used to describe an request to a server.
        :param start_response: Represents a response from a web request."""
        if not self.router:
            if environ.get('PATH_INFO') == '/':
                _status = '200 OK'
                _message = "Welcome to Retic!"
            else:
                _status = '404 Not found'
                _message = "error: The HTTP method {0} doesn't exist".format(
                    environ.get('REQUEST_METHOD')
                )
            start_response(_status, [('Content-Type', 'text/html')])
            return [_message.encode("utf8")]
        else:
            return self.router.main(environ, start_response)

    def clear(self):
        """Clear the App"""
        self.router: Router = None
        self.config.clear()
        self.apps.clear()

    def use(self, item: any, name: str = ""):
        """method of configuring the middleware.

        :param item: item of specific type for specific settings in a app
        :return: instance of the parant type for the item
        """

        """TODO: implement another types of item"""
        if isinstance(item, Router):
            self.router = item
        elif name:
            self.apps.setdefault(name, item)
        else:
            raise KeyError("error: A name for the item is necesary")

    def listen(
        self,
        hostname=APP_HOSTNAME,
        port=APP_PORT,
        application=None,
        use_reloader=False,
        use_debugger=False,
        use_evalex=True,
        extra_files=None,
        reloader_interval=1,
        reloader_type='auto',
        threaded=False,
        processes=1,
        request_handler=None,
        static_files=None,
        passthrough_errors=False,
        ssl_context=None
    ):
        """Create a server based in settings parameters.

        :param hostname: The host to bind to, for example ``'localhost'``.
            If the value is a path that starts with ``unix://`` it will bind
            to a Unix socket instead of a TCP socket..
        :param port: The port for the server.  eg: ``8080``
        :param application: the WSGI application to execute
        :param use_reloader: should the server automatically restart the python
                            process if modules were changed?
        :param use_debugger: should the werkzeug debugging system be used?
        :param use_evalex: should the exception evaluation feature be enabled?
        :param extra_files: a list of files the reloader should watch
                            additionally to the modules.  For example configuration
                            files.
        :param reloader_interval: the interval for the reloader in seconds.
        :param reloader_type: the type of reloader to use.  The default is
                            auto detection.  Valid values are ``'stat'`` and
                            ``'watchdog'``. See :ref:`reloader` for more
                            information.
        :param threaded: should the process handle each request in a separate
                        thread?
        :param processes: if greater than 1 then handle each request in a new process
                        up to this maximum number of concurrent processes.
        :param request_handler: optional parameter that can be used to replace
                                the default one.  You can use this to replace it
                                with a different
                                :class:`~BaseHTTPServer.BaseHTTPRequestHandler`
                                subclass.
        :param static_files: a list or dict of paths for static files.  This works
                            exactly like :class:`SharedDataMiddleware`, it's actually
                            just wrapping the application in that middleware before
                            serving.
        :param passthrough_errors: set this to `True` to disable the error catching.
                                This means that the server will die on errors but
                                it can be useful to hook debuggers in (pdb etc.)
        :param ssl_context: an SSL context for the connection. Either an
                            :class:`ssl.SSLContext`, a tuple in the form
                            ``(cert_file, pkey_file)``, the string ``'adhoc'`` if
                            the server should automatically create one, or ``None``
                            to disable SSL (which is the default).
        source by: werkzeug.serving
        """

        """TODO: Welcome message after the server is created"""
        run_simple(
            hostname=hostname,
            port=int(port),
            application=application or self.application,
            use_reloader=use_reloader,
            use_debugger=use_debugger,
            use_evalex=use_evalex,
            extra_files=extra_files,
            reloader_interval=reloader_interval,
            reloader_type=reloader_type,
            threaded=threaded,
            processes=processes,
            request_handler=request_handler,
            static_files=static_files,
            passthrough_errors=passthrough_errors,
            ssl_context=ssl_context
        )


class Config(object):
    def __init__(self, env):
        """Variables from the environment"""
        self.env = env
        """variables from user settings"""
        self.config = {}

    @property
    def env(self):
        return self.__env

    @env.setter
    def env(self, value):
        self.__env = value

    @property
    def config(self):
        return self.__config

    @config.setter
    def config(self, value):
        """If the type of the value is not dict, it is not allowed"""
        if not isinstance(value, dict):
            raise TypeError(
                "error: A settings dictionary of type dict is necesary"
            )
        self.__config = value

    def get(self, key, default_value=None, callback=None):
        """Returns the value of the parameter with the specified name.
        If the variable doesn't exist in the configuration values, this search
        in the environment variables and return a string. If you need a specific
        type of the environment variable, you need to use *app.env.int("variable_name")* for example.

        :param key: Name of the variable to find
        :param default_value: Value of the variable if this one doesn't exist
        :param callback: Function that is executed after getting the value
        """
        _value = self.__config.get(key, self.env(key, default_value))
        if not callback:
            return _value
        return callback(_value)

    def set(self, key, value):
        """Set a value in the settings of the app.

        Please note that names are not case sensitive.

        :param key: Name of the variable to set
        :param value: Value of the variable
        """
        self.__config.setdefault(key, value)

    def from_object(self, settings):
        """Set settings in based a dictionary, for example, if you want to
        set a additional configuration you nedd pass:

        ``app.config.from_object( { u'port': 8080 } )``

        :param settings: An object of type dictionary that contains the configurations
        """
        if not isinstance(settings, dict):
            raise TypeError(
                "error: A settings dictionary of type dict is necesary"
            )
        self.__config = {**self.__config, **settings}

    def clear(self):
        """Clear the actual settings, however, the settings from the environment 
        variables isn't clear. You can search variables in the environment with the function
        ``app.config.get("environment_name")``
        """
        self.__config.clear()



# System
from retic.lib.hooks.system.env import env

APP_LANG_LOCALEDIR = env("APP_LANG_LOCALEDIR", None)
APP_LANG_DOMAIN = env.str("APP_LANG_DOMAIN", "base")
APP_LANG_LANGUAGES = env.list("APP_LANG_LANGUAGES", ["en_US"])
APP_LANG_FALLBACK = env.bool("APP_LANG_FALLBACK", True)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='retic',
    version='0.0.15',
    license='MIT',
    description='Fastest, easiest and simple web framework for Python.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Braian Staimer',
    author_email='braianflorian@gmail.com',
    url='https://github.com/reticpy/retic',
    packages=setuptools.find_packages(),
    include_package_data=True,    
    download_url='https://github.com/reticpy/retic/archive/0.0.15.tar.gz',
    keywords=['PYTHON', 'WEB FRAMEWORK', 'MACHINE LEARNING WEB FRAMEWORK'],
    install_requires=[
        "environs>=7.4.0",
        "httpmethods>=1.0.5",
        "Werkzeug>=1.0.0",
        "repath>=0.9.0"
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



"""Validation services"""

def validate_obligate_fields(fields: any = None):
    """Validate if a list of obligate params are valid    

    :param fields: object that contains all params that are obligate, 
    these values can be arrays or simple values."""

    '''You can use the following example:

    _validate = validate_obligate_fields({     
        u'files': req.files.get("files", None)         
    })

    if _validate["valid"] is False:        
        return res.bad_request(            
            error_response_service(                
                "The param {} is necesary.".format(_validate["error"])                
            )            
        )
    '''
    if not fields:
        raise ValueError("error: A value for fields is necessary.")

    for field in fields:
        _item = fields.get(field, None)
        if _item == None \
                or (isinstance(_item, str) and _item == "") \
                or (isinstance(_item, list) and not _item):
            return {
                u'valid': False,
                u'error': field
            }
    return {
        u'valid': True,
        u'error': None
    }



# Json
from json import dumps, loads

# Retic
from retic.services.exceptions import get_file_error_exception


def jsonify(object: any):
    """Convert a object to a JSON string.

    :param object: is the client response object, if the object is str, 
    it returns this value, otherwise it creates a jsonify of the object
    """
    try:
        # if is none, return a empty string
        if not object:
            return ""
        # if is a bytesm return the same value
        elif isinstance(object, bytes):
            return object
        # if is a string, return the same value
        elif isinstance(object, str):
            return object
        # return the value in string format
        elif isinstance(object, dict):
            return dumps(object)
        # return a error message
        else:
            raise "error: The format of the object for the response is invalid."
    except Exception as e:
        return dumps({
            u"path": get_file_error_exception(3),
            u"error": str(e)
        })


def parse(content: str):
    """Deserialize (a str, bytes, or bytearray instance that contains 
    a JSON document) to a Python object.

    :param content: Content of type str, bytes, or bytearray that contains a valid JSON.
    """
    return loads(content)



METHODS = 'GET,POST,DELETE,PUT,OPTIONS'
HAS_CREDENTIALS = True
ORIGIN = "*"
HEADERS = "Content-Type"
EXPOSE_HEADERS = "Content-Type"


def cors(
    methods: str = METHODS,
    has_credentials: bool = HAS_CREDENTIALS,
    origin: str = ORIGIN,
    headers: str = HEADERS,
    expose_headers: str = EXPOSE_HEADERS
):
    """CORS is an abbreviation that stands for 'Cross-Origin Resource 
    Sharing' and, as the name implies, allows sharing of resources 
    from a variety of sources.

    A simple cross-origin request would be when domain1.com would be 
    accessing a resource from domain2.com (the resource is an image, 
    a CSS file, or something else). This has some massive security 
    implications, of course, and the built-in behavior for browsers 
    is that they would restrict the cross-origin HTTP request.

    :param methods: The Access-Control-Allow-Methods response header indicates 
    which HTTP methods are allowed on a particular endpoint for cross-origin requests.
    :param has_credentials: Access-Control-Allow-Credentials response header tells 
    the browser that the server allows 
    credentials for a cross-origin request.
    :param origin: The Access-Control-Allow-Origin response header indicates 
    whether the resources in the response can be shared with the given origin.
    :param headers: The Access-Control-Allow-Headers response header is used 
    in response to a preflight request that includes the Access-Control-Request-Headers 
    to indicate which HTTP headers can be used during the actual request
    :param expose_headers: The Access-Control-Expose-Headers response header indicates 
    which headers can be exposed as part of the response by listing

    ``from retic import Router``

    ``router = Router()``

    ``router.use(cors())``
    """

    def set_middleware(req, res, next):
        """Set all cors headers in a response to Client.

        :param req: Request is used to describe an request to a server.
        :param res: Represents a response from a web request.
        :param next: It must call next() to pass control to the next middleware function
        """

        res.set_headers('Access-Control-Allow-Methods', has_credentials)
        res.set_headers('Access-Control-Allow-Credentials', methods)
        res.set_headers('Access-Control-Allow-Origin', origin)
        res.set_headers('Access-Control-Allow-Headers', headers)
        res.set_headers('Access-Control-Expose-Headers', expose_headers)
        next()
    return set_middleware
